float4x4 World;
float4x4 View;
float4x4 Projection;
float4 Color;
float4 Tint;

texture Texture;
texture ShadowMap;
texture LightMap;

sampler colorSampler = sampler_state
{
    Texture = (Texture);
    MAGFILTER = POINT;
    MINFILTER = POINT;
    MIPFILTER = POINT;
    AddressU = Wrap;
    AddressV = Wrap;
};

sampler colorSamplerClamped = sampler_state
{
    Texture = (Texture);
    MAGFILTER = LINEAR;
    MINFILTER = LINEAR;
    MIPFILTER = LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};


sampler shadowSampler = sampler_state
{
    Texture = (ShadowMap);
    MAGFILTER = LINEAR;
    MINFILTER = LINEAR;
    MIPFILTER = LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

sampler lightSampler = sampler_state
{
    Texture = (LightMap);
    MAGFILTER = LINEAR;
    MINFILTER = LINEAR;
    MIPFILTER = LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};


struct VertexShaderInput
{
	float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
};

struct VertexShaderOutput
{
	float4 hPosition : POSITION0;
    float2 TexCoord : TEXCOORD0;
	float2 WorldTexCoord : TEXCOORD1;
};

VertexShaderOutput VertexShaderFunction(VertexShaderInput input)
{
    VertexShaderOutput output;
    
    float4 worldPosition = mul(input.Position, World);
    float4 viewPosition = mul(worldPosition, View);
    output.hPosition = mul(viewPosition, Projection);

    output.TexCoord = input.TexCoord; 
	output.WorldTexCoord = float2((output.hPosition.x / 2) + 0.5, 1.0 - ( (output.hPosition.y / 2) + 0.5));
    return output;
}

struct PixelShaderOutput
{
    float4 Color : COLOR0;
};

PixelShaderOutput PSDrawShadow(VertexShaderOutput input)
{
	PixelShaderOutput output;
	//output.Color = float4(1.0f,1.0f,1.0f,1.0f);
	float4 ShadowColor = tex2D(colorSamplerClamped, input.TexCoord);
	output.Color = ShadowColor;
	//output.Color.a = output.Color.r;
	return output;
}

PixelShaderOutput PSDrawLight(VertexShaderOutput input)
{
	PixelShaderOutput output;
	float4 TColor = tex2D(colorSampler, input.TexCoord) / 4;
	float4 ShadowColor = tex2D(shadowSampler, input.WorldTexCoord);
	output.Color.a = TColor.r;
	output.Color.r = Color.r * TColor.r * TColor.a * (1.0 - ShadowColor.r);
	output.Color.g = Color.g * TColor.r * TColor.a * (1.0 - ShadowColor.r);
	output.Color.b = Color.b * TColor.r * TColor.a * (1.0 - ShadowColor.r);
	return output;
}

PixelShaderOutput PSDrawLight2(VertexShaderOutput input)
{
	PixelShaderOutput output;
	float4 TColor = tex2D(colorSampler, input.TexCoord) / 4;
	output.Color.a = TColor.r;
	output.Color.r = Color.r * TColor.r * TColor.a;
	output.Color.g = Color.g * TColor.r * TColor.a;
	output.Color.b = Color.b * TColor.r * TColor.a;
	return output;
}

PixelShaderOutput PSDrawLighted(VertexShaderOutput input)
{
	PixelShaderOutput output;
	output.Color = Tint * Color * tex2D(colorSampler, input.TexCoord) * (tex2D(lightSampler, input.WorldTexCoord) * 4);
	output.Color.r *= output.Color.a;
	output.Color.g *= output.Color.a;
	output.Color.b *= output.Color.a;
	return output;
}

PixelShaderOutput PSDrawFullbright(VertexShaderOutput input)
{
	PixelShaderOutput output;
	output.Color = Tint * Color * tex2D(colorSampler, input.TexCoord);
	output.Color.r *= output.Color.a;
	output.Color.g *= output.Color.a;
	output.Color.b *= output.Color.a;
	return output;
}

PixelShaderOutput PSDrawFullbrightClamped(VertexShaderOutput input)
{
	PixelShaderOutput output;
	output.Color = Tint * Color * tex2D(colorSamplerClamped, input.TexCoord);
	output.Color.r *= output.Color.a;
	output.Color.g *= output.Color.a;
	output.Color.b *= output.Color.a;
	return output;
}

technique DrawShadow
{
    pass Pass1
    {
		AlphaBlendEnable = true;
		BlendOp = Add;
		SrcBlend = One;
		DestBlend = One;
		ZEnable = false;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader = compile ps_2_0 PSDrawShadow();
    }
}

technique DrawLight
{
    pass Pass1
    {
		AlphaBlendEnable = true;
		BlendOp = Add;
		SrcBlend = One;
		DestBlend = One;
		ZEnable = false;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader = compile ps_2_0 PSDrawLight();
    }
}

technique DrawLight2
{
    pass Pass1
    {
		AlphaBlendEnable = true;
		BlendOp = Add;
		SrcBlend = One;
		DestBlend = One;
		ZEnable = false;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader = compile ps_2_0 PSDrawLight2();
    }
}

technique DrawLighted
{
    pass Pass1
    {
		AlphaBlendEnable = true;
		BlendOp = Add;
		SrcBlend = One;
		DestBlend = InvSrcAlpha;
		ZEnable = false;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader = compile ps_2_0 PSDrawLighted();
    }
}

technique DrawFullbright
{
    pass Pass1
    {
		AlphaBlendEnable = true;
		BlendOp = Add;
		SrcBlend = One;
		DestBlend = InvSrcAlpha;
		ZEnable = false;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader = compile ps_2_0 PSDrawFullbright();
    }
}

technique DrawFullbrightClamped
{
    pass Pass1
    {
		AlphaBlendEnable = true;
		BlendOp = Add;
		SrcBlend = One;
		DestBlend = InvSrcAlpha;
		ZEnable = false;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader = compile ps_2_0 PSDrawFullbrightClamped();
    }
}