﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using System.ComponentModel;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace Jemgine.Data
{
    [Jemgine.Engine.Category("Primitive"), Jemgine.Engine.Component]
    public partial class Quad : EntityElement, EditorDrawable, TexturedObject,
        Translatable, Rotatable, Scalable, ReferenceTarget, Tintable, 
        HasVerticies, Engine.Renderer.IDrawable
    {
        public int BoneID { get; set; }
        [ContentSerializerIgnore]
        public Bone BoneReference { get; set; }

        public Vector4 Color { get; set; }

        [Microsoft.Xna.Framework.Content.ContentSerializerIgnore]
        public int ReferenceID { get; set; }

        public Vector2 GetCenter()
        {
            Vector2 Center = Vector2.Zero;
            foreach (var V in Data)
                Center += V.Position;
            return Center / Data.Length;
        }

        Vector2 ReferenceTarget.GetWorldPosition()
        {
            if (Owner == null) return Vector2.Zero;
            else return Owner.Position + GetCenter();
        }

        [Microsoft.Xna.Framework.Content.ContentSerializerIgnore]
        public EntityRecord Owner { get; private set; }

        public void SetOwner(EntityRecord E)
        {
            Owner = E;
        }

        public Quad(Quad p)
        {
            for (int i = 0; i < 4; ++i) 
                Data[i] = new Vertex(p.Data[i].Position);
            Texture = p.Texture;
            Color = p.Color;
            BoneID = p.BoneID;

        }

        public Quad()
        {
            Color = Vector4.One;
            for (int i = 0; i < 4; ++i) Data[i] = new Vertex();
        }

        public Vertex[] Data = new Vertex[4];

        [System.ComponentModel.Editor(typeof(TextureEditor), typeof(System.Drawing.Design.UITypeEditor))]
        public String Texture { get; set; }

        public void ClearDecomposedCache() { }
      
        public int Count { get { return 4; } }
        public Vertex this[int _index]
        {
            get { return Data[_index]; }
            set { Data[_index] = value; }
        }
        public IVertex GetVertex(int Index)
        {
            return this[Index] as IVertex;
        }

        public bool PointInside(Vector2 point)
        {
            double angle = 0;

            for (int i = 0; i < Count; i++)
            {
                Vector2 p1 = this[i].Position - point;
                Vector2 p2 = this[i == Count - 1 ? 0 : i + 1].Position - point;

                angle += Math.Vector.AngleBetweenVectors(p1, p2);
            }

            if (System.Math.Abs(angle) < System.Math.PI - Math.Utility.Epsilon)
                return false;

            return true;
        }

        public Jemgine.Engine.Selection.Types SelectionType()
        {
            return Jemgine.Engine.Selection.Types.EntityElement;
        }

        public EntityElement Clone()
        {
            return new Quad(this);
        }

        public void Initialize(Data.EntityRecord Self, Microsoft.Xna.Framework.Content.ContentManager Content)
        { }

        public void Destroy()
        { }

        #region Translatable Members

        object Translatable.PrepareToTranslate()
        {
            List<Vector2> Positions = new List<Vector2>();
            foreach (var Vertex in Data) Positions.Add(Vertex.Position);
            return Positions;
        }

        void Translatable.ApplyTranslation(object Tag, Vector2 MouseDelta)
        {
            List<Vector2> Positions = Tag as List<Vector2>;
            for (int i = 0; i < Data.Length; ++i)
                Data[i].Position = Positions[i] + MouseDelta;
            ClearDecomposedCache();
        }

        void Translatable.Snap(Snapper Snapper)
        {
            foreach (var Vertex in Data)
                Vertex.Position = Snapper.SnapPosition(Vertex.Position);
        }

        #endregion

        #region Rotatable Members

        object Rotatable.PrepareToRotate()
        {
            List<Vector2> Positions = new List<Vector2>();
            foreach (var Vertex in Data) Positions.Add(Vertex.Position);
            return Positions;
        }

        void Rotatable.ApplyRotation(object Tag, float Rotation)
        {
            List<Vector2> Positions = Tag as List<Vector2>;
            Vector2 Center = Vector2.Zero;
            foreach (var V in Positions) Center += V;
            Center /= Positions.Count;

            Matrix M = Matrix.CreateRotationZ(Rotation);

            for (int I = 0; I < Positions.Count; ++I)
                Data[I].Position = Center + Vector2.Transform(Positions[I] - Center, M);

            ClearDecomposedCache();
        }

        void Rotatable.Snap(Snapper Snapper)
        {
            (this as Translatable).Snap(Snapper);
        }

        #endregion

        #region EditorDrawable Members

        public void Draw(int ID, EditorDrawData Context, SelectionType Selected)
        {
            Context.Renderer.Color = EditorDrawData.DrawColor(Selected);

            var _Texture = Context.GetTexture(Texture);
            if (Selected != Jemgine.Data.SelectionType.NotSelected)
                _Texture = null;

            Context.Renderer.DrawQuad(this, _Texture);

            if (Selected == Jemgine.Data.SelectionType.Selected)
                foreach (Vertex point in Data)
                    Context.Renderer.DrawPoint(point.Position);
        }

        #endregion

        #region TexturedObject Members

        public void SetTexture(String texture, Texture2D _tex)
        {
            this.Texture = texture;
        }

        #endregion

        #region Scalable Members

        object Scalable.PrepareToScale()
        {
            List<Vector2> Positions = new List<Vector2>();
            foreach (var Vertex in Data) Positions.Add(Vertex.Position);
            return Positions;
        }

        void Scalable.ApplyScale(object Tag, float Delta)
        {
            List<Vector2> Positions = Tag as List<Vector2>;
            Vector2 Center = Vector2.Zero;
            foreach (var V in Positions) Center += V;
            Center /= Positions.Count;

            float MinY = Positions[0].Y;
            float MaxY = Positions[0].Y;
            foreach (var V in Positions)
            {
                if (V.Y > MaxY) MaxY = V.Y;
                if (V.Y < MinY) MinY = V.Y;
            }

            float Range = MaxY - MinY;
            float NewRange = Range + Delta;

            Matrix M = Matrix.CreateScale(NewRange / Range);

            for (int I = 0; I < Positions.Count; ++I)
                Data[I].Position = Center + Vector2.Transform(Positions[I] - Center, M);

            ClearDecomposedCache();
        }

        void Scalable.Snap(Snapper Snapper)
        {
            (this as Translatable).Snap(Snapper);
        }

        #endregion

        public void SetTint(Vector4 Tint)
        {
            Color = Tint;
        }

        public Vector4 GetTint()
        {
            return Color;
        }


        void Engine.Renderer.IDrawable.Draw(
            Engine.Renderer.RenderEffect Effect, 
            Engine.Renderer.VertexPusher Pusher, 
            Engine.EpisodeContentManager Content)
        {
            Effect.Texture = Content.GetTexture(Texture);
            Effect.Color = Color;

            Pusher.VertexBuffer[0].Position = new Vector3(Data[0].Position, 0.0f);
            Pusher.VertexBuffer[1].Position = new Vector3(Data[1].Position, 0.0f);
            Pusher.VertexBuffer[3].Position = new Vector3(Data[2].Position, 0.0f);
            Pusher.VertexBuffer[2].Position = new Vector3(Data[3].Position, 0.0f);
            Pusher.VertexBuffer[0].TextureCoordinate = new Vector2(0, 1);
            Pusher.VertexBuffer[1].TextureCoordinate = new Vector2(0, 0);
            Pusher.VertexBuffer[3].TextureCoordinate = new Vector2(1, 0);
            Pusher.VertexBuffer[2].TextureCoordinate = new Vector2(1, 1);

            Pusher.DrawTriangleStrip(Pusher.VertexBuffer);
        //    if (DecomposedRuntimePolygons == null ||
        //        (DecomposedRuntimePolygons.Count >= 1 && DecomposedRuntimePolygons[0].VertexCache == null))
        //        UpdateDecomposedCache();
        //    foreach (var Poly in DecomposedRuntimePolygons)
        //        Pusher.DrawTriangleStrip(Poly.VertexCache);
        }
    }
}
